use super::{NPCMessageHandler, NPCSayInfo, NPCSayType, NPCSpawnPossession, NPC};
#[double]
use crate::db::DBTrans;
use crate::{
    message_handler::user_commands::{UResult, VerbContext},
    models::{
        consent::ConsentType,
        item::{Item, ItemFlag, ItemSpecialData, LocationActionType, Pronouns, SkillType},
        task::{Task, TaskDetails, TaskMeta},
    },
    regular_tasks::{TaskHandler, TaskRunContext},
    services::comms::broadcast_to_room,
    static_content::{
        npc::KillBonus, possession_type::PossessionType, room::Direction, species::SpeciesType,
    },
    DResult,
};
use ansi::ansi;
use async_trait::async_trait;
use chrono::Utc;
use itertools::Itertools;
use mockall_double::double;
use nom::{
    bytes::complete::tag,
    character::complete::{multispace1, u8},
    combinator::{eof, opt},
    sequence::{delimited, pair, preceded, terminated},
};
use std::time;

async fn reply(ctx: &VerbContext<'_>, msg: &str) -> DResult<()> {
    ctx.trans
        .queue_for_session(
            &ctx.session,
            Some(&format!(
                ansi!("Doorbot beeps, then replies in low-fidelity audio: <yellow>{}<reset>\n"),
                msg
            )),
        )
        .await
}

fn parse_move_message(mut input: &str) -> Result<(u8, u8), &str> {
    input = input.trim();
    let (from, to) = match terminated(
        preceded(
            preceded(
                tag("move"),
                preceded(
                    multispace1::<&str, ()>,
                    opt(preceded(tag("from"), multispace1::<&str, ()>)),
                ),
            ),
            pair(
                u8,
                preceded(delimited(multispace1, tag("to"), multispace1), u8),
            ),
        ),
        eof,
    )(input)
    {
        Ok((_, v)) => Ok(v),
        Err(_) => Err(
            "Invalid command, feeble human. I only understand commands like: -doorbot move 1 to 2",
        ),
    }?;
    if from == 0 || to == 0 {
        return Err("I like starting numbers at 0 too, but, alas, my programmer was a feeble human and numbered the first tower 1!");
    }
    if from > 3 || to > 3 {
        return Err(
            "Feeble human, this will make it hard for your simple brain, but I have but 3 towers!",
        );
    }
    if from == to {
        return Err(
            "Only a feeble human would try to move a disc from a tower... to the same tower!",
        );
    }
    Ok((from, to))
}

type TowerData = [Vec<u8>; 3];
fn move_between_towers(from: u8, to: u8, initial: &TowerData) -> Result<TowerData, &'static str> {
    let mut new_dat = initial.clone();
    let move_disc = new_dat[(from - 1) as usize]
        .pop()
        .ok_or_else(|| "Feeble human, you can't move a disc from a tower with no discs.")?;
    match new_dat[(to - 1) as usize].last() {
        Some(onto_disc) if onto_disc > &move_disc => {
            Err("Feeble human, the rules of the game say you must never place a larger disc on top of a smaller one.")?;
        }
        _ => {}
    }
    new_dat[(to - 1) as usize].push(move_disc);
    return Ok(new_dat);
}

fn describe_towers(data: &TowerData) -> String {
    let mut description = String::new();
    for (discs, word) in data.iter().zip(["first", "second", "third"]) {
        description.push_str(&format!("The {} tower ", word));
        if discs.len() == 0 {
            description.push_str("is empty. ");
        } else {
            description.push_str(&format!(
                "holds {}. ",
                discs
                    .iter()
                    .rev()
                    .map(|d| match d {
                        4 => "a very small red disc",
                        3 => "a small orange disc",
                        2 => "a medium-sized yellow disc",
                        _ => "a large green disc",
                    })
                    .join(", atop ")
            ));
        }
    }
    description + "A cursor blinks as if waiting for further input"
}

fn is_solved(data: &TowerData) -> bool {
    data[2].len() == 4
}

fn initial_towers() -> TowerData {
    [vec![1, 2, 3, 4], vec![], vec![]]
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn parse_move_message_accepts_valid() {
        assert_eq!(parse_move_message(" move  1 to  2"), Ok((1, 2)));
    }

    #[test]
    fn parse_move_message_accepts_valid_with_from() {
        assert_eq!(parse_move_message(" move from 1 to  2"), Ok((1, 2)));
    }

    #[test]
    fn parse_move_message_rejects_badstart() {
        assert_eq!(parse_move_message("eat  1 to  2"), Err("Invalid command, feeble human. I only understand commands like: -doorbot move 1 to 2"));
    }

    #[test]
    fn parse_move_message_rejects_negative() {
        assert_eq!(parse_move_message("move -1 to  2"), Err("Invalid command, feeble human. I only understand commands like: -doorbot move 1 to 2"));
    }

    #[test]
    fn parse_move_message_rejects_zero() {
        assert_eq!(parse_move_message("move 1 to 0"), Err("I like starting numbers at 0 too, but, alas, my programmer was a feeble human and numbered the first tower 1!"));
    }

    #[test]
    fn parse_move_message_rejects_four() {
        assert_eq!(parse_move_message("move 4 to 1"), Err("Feeble human, this will make it hard for your simple brain, but I have but 3 towers!"));
    }

    #[test]
    fn parse_move_message_rejects_self() {
        assert_eq!(
            parse_move_message("move 2 to 2"),
            Err("Only a feeble human would try to move a disc from a tower... to the same tower!")
        );
    }

    #[test]
    fn move_between_towers_works_to_empty_tower() {
        assert_eq!(
            move_between_towers(1, 2, &[vec![2, 3, 4], vec![], vec![1]]),
            Ok([vec![2, 3], vec![4], vec![1]])
        )
    }

    #[test]
    fn move_between_towers_works_to_nonempty_tower() {
        assert_eq!(
            move_between_towers(1, 3, &[vec![4], vec![], vec![1, 2, 3]]),
            Ok([vec![], vec![], vec![1, 2, 3, 4]])
        )
    }

    #[test]
    fn move_between_towers_fails_from_empty_tower() {
        assert_eq!(
            move_between_towers(2, 3, &initial_towers()),
            Err("Feeble human, you can't move a disc from a tower with no discs.")
        )
    }

    #[test]
    fn describe_towers_works() {
        assert_eq!(
            describe_towers(&[vec![1,2,3,4], vec![], vec![]]),
            "The first tower holds a very small red disc, atop a small orange disc, atop a medium-sized yellow disc, atop a large green disc. \
             The second tower is empty. \
             The third tower is empty. A cursor blinks as if waiting for further input"
        )
    }

    #[test]
    fn is_solved_returns_true_when_solved() {
        assert_eq!(is_solved(&[vec![], vec![], vec![1, 2, 3, 4]]), true)
    }
    #[test]
    fn is_solved_returns_false_when_not_solved() {
        assert_eq!(is_solved(&[vec![], vec![4], vec![1, 2, 3]]), false)
    }
}

struct DoorbotMsgHandler;
#[async_trait]
impl NPCMessageHandler for DoorbotMsgHandler {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _source: &Item,
        target: &Item,
        mut message: &str,
    ) -> UResult<()> {
        message = message.trim();
        let (from, to) = match parse_move_message(message) {
            Err(e) => {
                reply(ctx, e).await?;
                return Ok(());
            }
            Ok((from, to)) => (from, to),
        };

        let initial_towerdata = match target.special_data {
            Some(ItemSpecialData::HanoiPuzzle { towers: ref v }) => v.clone(),
            _ => initial_towers(),
        };

        let new_towerdata = match move_between_towers(from, to, &initial_towerdata) {
            Err(e) => {
                reply(ctx, e).await?;
                return Ok(());
            }
            Ok(d) => d,
        };

        if is_solved(&new_towerdata) {
            reset_puzzle(&ctx.trans, true).await?;
            let msg = ansi!("Doorbot moves a disc to the third tower, completing the puzzle. \
                             Doorbot beeps and says: <blue>\"I underestimated you, mighty human. \
                             You have proven your worth to pass through the door to the hacker's \
                             club. I will open the door for you - when it closes it will be locked \
                             again, ready for the next candidate for the club - so I suggest you enter \
                             before it closes.\"<reset>. As if by magic, the puzzle resets back to \
                             its starting state, and the door swings open with a creak.\n");
            broadcast_to_room(&ctx.trans, &target.location, None, msg, Some(msg)).await?;
            ctx.trans
                .upsert_task(&Task {
                    meta: TaskMeta {
                        task_code: "room/computer_museum_hackers_club/south".to_owned(),
                        next_scheduled: Utc::now() + chrono::Duration::seconds(120),
                        ..Default::default()
                    },
                    details: TaskDetails::SwingShut {
                        room_item: "room/computer_museum_hackers_club".to_owned(),
                        direction: Direction::SOUTH,
                    },
                })
                .await?;
        } else {
            let mut room_item = (*ctx
                .trans
                .find_item_by_type_code("room", "computer_museum_club_door")
                .await?
                .ok_or_else(|| "Missing room computer_musuem_club_door")?)
            .clone();
            let descr = describe_towers(&new_towerdata);
            let msg = format!(
                "Doorbot moves a disc to tower {}. The screen now shows: {}.\n",
                to, &descr
            );
            broadcast_to_room(&ctx.trans, &target.location, None, &msg, Some(&msg)).await?;
            room_item.details_dyn_suffix = Some(descr);
            ctx.trans.save_item_model(&room_item).await?;

            ctx.trans
                .upsert_task(&Task {
                    meta: TaskMeta {
                        task_code: "reset_hanoi_computer_club".to_owned(),
                        next_scheduled: Utc::now() + chrono::Duration::seconds(600),
                        ..Default::default()
                    },
                    details: TaskDetails::ResetHanoi,
                })
                .await?;
            let mut target_mut = (*target).clone();
            target_mut.special_data = Some(ItemSpecialData::HanoiPuzzle {
                towers: new_towerdata,
            });
            ctx.trans.save_item_model(&target_mut).await?;
        }

        Ok(())
    }
}

async fn reset_puzzle(trans: &DBTrans, door_state: bool) -> DResult<()> {
    let mut npc_item = (*trans
        .find_item_by_type_code("npc", "computer_museum_doorbot")
        .await?
        .ok_or_else(|| "Missing npc computer_musuem_doorbot")?)
    .clone();
    npc_item.special_data = Some(ItemSpecialData::HanoiPuzzle {
        towers: initial_towers(),
    });
    trans.save_item_model(&npc_item).await?;
    let mut doorwell_room_item = (*trans
        .find_item_by_type_code("room", "computer_museum_club_door")
        .await?
        .ok_or_else(|| "Missing room computer_musuem_club_door")?)
    .clone();
    let mut club_room_item = (*trans
        .find_item_by_type_code("room", "computer_museum_hackers_club")
        .await?
        .ok_or_else(|| "Missing room computer_musuem_club_door")?)
    .clone();
    match club_room_item.door_states {
        None => {}
        Some(ref mut d) => match d.get_mut(&Direction::SOUTH) {
            None => {}
            Some(d) => d.open = door_state,
        },
    }
    doorwell_room_item.details_dyn_suffix = Some(describe_towers(&initial_towers()));
    trans.save_item_model(&doorwell_room_item).await?;
    trans.save_item_model(&club_room_item).await?;
    Ok(())
}

pub struct ResetGameTaskHandler;
#[async_trait]
impl TaskHandler for ResetGameTaskHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        reset_puzzle(&ctx.trans, false).await?;
        Ok(None)
    }
}
pub static RESET_GAME_HANDLER: &'static (dyn TaskHandler + Sync + Send) = &ResetGameTaskHandler;

pub fn npc_list() -> Vec<NPC> {
    use NPCSayType::FromFixedList;
    let geek_stdsay = NPCSayInfo {
        say_code: "babble",
        frequency_secs: 120,
        talk_type: FromFixedList(vec!(
            (false, "I wish I could get into the secret hackers club - I heard they have a machine that hacks your wristpad to make you a super smart dork."),
            (false, "The basement here is pretty dangerous with all those robots."),
            (false, "I overhead a real dork bragging about how he managed to get all four rings onto the rightmost virtual peg. No idea what that was about!"),
            (false, "Why is it that all the dorks around here keep whispering stuff to each other about Towers of Hanoi?"),
        ))
    };

    let killbot_stdsay = NPCSayInfo {
        say_code: "babble",
        frequency_secs: 20,
        talk_type: FromFixedList(vec!(
            (false, "My mission is to EXTERMINATE"),
            (false, "Intruder kill mode active"),
            (false, "Recoverable Error: Preset Kill Limit not set. Treating as unlimited"),
            (false, "404 action module STUN not found. Falling back to action KILL underscore WITH underscore EXTREME underscore PREJUDICE"),
        ))
    };

    macro_rules! geek_gender_word {
        (default_male) => {
            "guy"
        };
        (default_female) => {
            "gal"
        };
    }

    macro_rules! geek {
        ($code: expr, $adj: expr, $spawn: expr, $pronouns: ident) => {
            NPC {
                code: concat!("computer_museum_geek_", $code),
                name: concat!($adj, " geek"),
                pronouns: Pronouns { is_proper: false, ..Pronouns::$pronouns() },
                description: concat!("A geeky looking ", geek_gender_word!($pronouns), " with horn-rimmed glasses, and a button-down shirt. In the pocket is a pocket protector"),
                spawn_location: concat!("room/computer_museum_", $spawn),
                spawn_possessions: vec![
                    NPCSpawnPossession {
                        what: PossessionType::Shirt,
                        action_type: LocationActionType::Worn,
                        wear_layer: 0,
                    },
                    NPCSpawnPossession {
                        what: PossessionType::Jeans,
                        action_type: LocationActionType::Worn,
                        wear_layer: 0,
                    },
                ],
                message_handler: None,
                wander_zones: vec!("computer_museum"),
                says: vec!(geek_stdsay.clone()),
                player_consents: vec!(ConsentType::Medicine),
                ..Default::default()
            }
        }
    }

    macro_rules! killbot {
        ($code: expr, $adj: expr, $spawn: expr) => {
            NPC {
                code: concat!("computer_museum_killbot_", $code),
                name: concat!($adj, " Killbot"),
                species: SpeciesType::Robot,
                pronouns: Pronouns {
                    is_proper: false,
                    ..Pronouns::default_inanimate()
                },
                description: concat!("A mean looking robot, complete with a vicious spiked mace and powerful solonoids for smashing it hard into whatever unfortunate victim this bot targets"),
                spawn_location: concat!("room/computer_museum_", $spawn),
                spawn_possessions: vec![],
                aggression: 21,
                aggro_pc_only: true,
                total_xp: 3000,
                total_skills: SkillType::values()
                    .into_iter()
                    .map(|sk| {
                        (
                            sk.clone(),
                            if &sk == &SkillType::Clubs { 12.0 } else { 10.0 },
                        )
                    })
                    .collect(),
                intrinsic_weapon: Some(PossessionType::SpikedMace),
                message_handler: None,
                wander_zones: vec!["computer_museum"],
                kill_bonus: Some(KillBonus {
                    msg: "On your wristpad: Those robots have been the bane of our existence at the museum. Thanks for taking one out! Here's a small token of our gratitude.",
                    payment: 120,
                }),
                says: vec![killbot_stdsay.clone()],
                player_consents: vec![ConsentType::Fight],
                ..Default::default()
            }
        };
    }

    vec![
        geek!("1", "excited", "lobby", default_male),
        geek!("2", "mesmerised", "lobby", default_female),
        killbot!("3", "abhorrent", "hw_1"),
        killbot!("4", "berserk", "hw_2"),
        killbot!("5", "vicious", "hw_3"),
        killbot!("6", "murderous", "club_door"),
        NPC {
            code: "computer_museum_doorbot",
            name: "Doorbot",
            spawn_location: "room/computer_museum_club_door",
            description: ansi!(
                "A flat panel screen on the door, with a microphone listening for \
                 commands directed at it, apparently hooked up to the door lock. \
                 [Hint: try <bold>-doorbot move from 1 to 2<reset> to ask Doorbot \
                 to move the top disc from tower 1 to tower 2]"
            ),
            pronouns: Pronouns::default_inanimate(),
            species: SpeciesType::Robot,
            extra_flags: vec![ItemFlag::DontListInLook],
            message_handler: Some(&DoorbotMsgHandler),
            ..Default::default()
        },
    ]
}
